import std/[critbits, re, strutils, strmisc]

import bio
import itertools

import constants
import memory
import types

proc tovar(code: char): ArcticVariableIndex =
    case code.toLowerAscii:
        of 'a': VARIABLE_A
        of 'b': VARIABLE_B
        of 'c': VARIABLE_C
        of 'i': VARIABLE_I
        of 'n': VARIABLE_N
        of 'x': VARIABLE_X
        of 'y': VARIABLE_Y
        of 'z': VARIABLE_Z
        else: raise new ValueError

proc parse_immediate(imm: string, section: string): ArcticImmediate =
    if imm.match(re"^\d+$"):
        return ArcticImmediate(kind: INTEGER, i: imm.parseInt)
    elif imm.match(re"^[AaBbCcIiNnXxYyZz]$"):
        return ArcticImmediate(kind: VARIABLE, v: imm[0].tovar)
    elif imm.len > 0 and imm[0] in "0123456789.":
        return ArcticImmediate(kind: NUMBER, n: imm.parseFloat)
    else:
        if imm.len > 0 and imm[0] == '_':
            return ArcticImmediate(kind: SYMBOL, s: section & " " & imm)
        else:
            return ArcticImmediate(kind: SYMBOL, s: imm)

func grabnum(data: string): (string, string) =
    ## Returns the numeric literal at the start of data, then the rest of data
    var i = 0
    while i < data.len and data[i] in "0123456789.":
        result[0] &= data[i]
        i += 1

    result[1] = data[i..^1]

proc parse_data(data: string): seq[uint8] =
    if data.len == 0:
        return

    template insertInt(itype, utype, endianness) =
        let (text, rest) = data[1..^1].grabnum

        if text[0] == '-':
            let val: itype = itype(-text[1..^1].parseBiggestInt)
            result.add cast[utype](val).serialize(endianness)
        else:
            let val: utype = utype(text.parseBiggestUInt)
            result.add val.serialize(endianness)

        result.add parse_data(rest)

    case data[0]:
        of 'i': insertInt(int64, uint64, littleEndian)
        of 'I': insertInt(int64, uint64, bigEndian)
        of 'w': insertInt(int32, uint32, littleEndian)
        of 'W': insertInt(int32, uint32, bigEndian)
        of 's': insertInt(int16, uint16, littleEndian)
        of 'S': insertInt(int16, uint16, bigEndian)
        of 'b', 'B': # bytes don't have an endianness
            let (text, rest) = data[1..^1].grabnum

            if text[0] == '-':
                let val: int8 = int8(-text[1..^1].parseBiggestInt)
                result.add cast[uint8](val)
            else:
                let val: uint8 = uint8(text.parseBiggestUInt)
                result.add val
            result.add parse_data(rest)

        of 'f':
            let (text, rest) = data[1..^1].grabnum
            result.add cast[uint64](text.parseFloat).serialize(littleEndian)
            result.add parse_data(rest)

        of 'F':
            let (text, rest) = data[1..^1].grabnum
            result.add cast[uint64](text.parseFloat).serialize(bigEndian)
            result.add parse_data(rest)

        of 'x': # hexadecimal byte constant
            let (hex, _, rest) = data[1..^1].partition("x")
            for pair in hex.chunked(2):
                result.add uint8(parseHexInt(pair.join))
            result.add parse_data(rest)

        of '"': # utf-8 string constant with escapes
            let strlen = data.matchLen(re"^""([^""]|\\"")*""")
            assert strlen >= 0
            for c in unescape(data[1..strlen - 2], prefix="", suffix=""):
                result.add c.uint8
            result.add parse_data(data[strlen..^1])

        else:
            result.add parse_data(data[1..^1])

func iscode(secname: string): bool =
    if secname.len == 0:
        return true
    return secname[0] in "cC"

func isbigendian(secname: string): bool =
    if secname.len == 0:
        return false
    return secname[0] in "C"

func raw_op(code: char): ArcticOperation =
    result.code = code
    result.immediate = ArcticImmediate(kind: PLAIN)

func load*(code: string): ArcticState =
    var
        section: string = ""
        token: string = ""

    result.pc = 1

    for next in code:
        if token.len == 0: # initial state
            case next:
                of ImmediateOps, '#', ':', '\"', '\'':
                    token &= next
                of PlainOps, '\n':
                    result.code.add(section, raw_op(next))
                else:
                    discard
        else:
            case token[0]:
                of '#': # comment
                    if next == '\n':
                        token = ""
                        result.code.add(section, raw_op(next))

                of '"': # section switch
                    if next == '"' and token[^1] != '\\':
                        section = token[1..^1]
                        token = ""
                    else:
                        token.add next

                of ':': # label name
                    if next == ' ' or next == '\n':
                        let idx = if token[1] == '_': section & " " & token[1..^1] else: token[1..^1]
                        if section.iscode:
                            result.symbols[idx] = result.code.current(section)
                        else:
                            result.symbols[idx] = result.code.current(section)
                        token = ""
                    else:
                        token.add next

                of '\'': # data injection
                    if next == '\'':
                        var
                            quoted = false
                            escaped = false
                        for c in token:
                            if c == '"':
                                if not escaped:
                                    quoted = not quoted
                                escaped = false
                            elif c == '\\':
                                escaped = true
                            else:
                                escaped = false

                        if quoted or escaped:
                            token.add next
                        else:
                            result.memory.add(section, parse_data(token[1..^1]))
                            token = ""
                    else:
                        token.add next

                of ImmediateOps:
                    if next in " \n":
                        result.code.add(section, ArcticOperation(
                            code: token[0],
                            immediate: parse_immediate(token[1..^1], section),
                            bigendian: isbigendian(section)))
                        token = ""
                    else:
                        token.add next

                else:
                    discard

when isMainModule:
    let state = stdin.readAll.load
    echo state
