import std/[critbits, math, tables]

type
    Memory*[T] = object
        ## Memory holds a set of sections and maps addresses in those sections
        ## to the range (0, int.high) via a low discrepancy sequence (in this
        ## case, the van der corput sequence)
        chunks: seq[seq[T]]
        sects: CritBitTree[int] # which chunk is the section
        # TODO: keep track of lengths and "skip" vdc indices that would overlap
        #       (only an issue where e.g. section 0 is huge and there are lots
        #       of sections)
        smlmem: Table[int, seq[T]] # index -> buffer
        midmem: Table[int, seq[T]] # index -> buffer
        bigmem: Table[int, seq[T]] # index -> buffer
        hugmem: Table[int, seq[T]] # index -> buffer
        # dynamic addresses are:
        #  (1 shl 18)-element blocks x (1 shl 42) starting from (int.high shr 1)
        #  (1 shl 26)-element blocks x (1 shl 34) starting from (int.high shr 1) + (1 shl 60)
        #  (1 shl 34)-element blocks x (1 shl 26) starting from (int.high shr 1) + 2 * (1 shl 60)
        #  (1 shl 42)-element blocks x (1 shl 18) starting from (int.high shr 1) + 3 * (1 shl 60)

const
    HUG_BLOCK_SIZE = 1 shl 42
    BIG_BLOCK_SIZE = 1 shl 34
    MID_BLOCK_SIZE = 1 shl 26
    SML_BLOCK_SIZE = 1 shl 18

    HUG_BLOCK_COUNT = 1 shl 18
    BIG_BLOCK_COUNT = 1 shl 26
    MID_BLOCK_COUNT = 1 shl 34
    SML_BLOCK_COUNT = 1 shl 42

    SML_BLOCK_START = int.high shr 1
    MID_BLOCK_START = SML_BLOCK_START + (1 shl 60)
    BIG_BLOCK_START = MID_BLOCK_START + (1 shl 60)
    HUG_BLOCK_START = BIG_BLOCK_START + (1 shl 60)

proc van_der_corput(n: int): int =
    var
        q = 1 # 0 is reserved for null
        i = n
        b = int.high shr 2 # int.high/2 .. int.high is reserved for dynmem

    while i > 0:
        q += (i and 1) * b
        i = i shr 1
        b = b shr 1

    return q

proc add*[T](memory: var Memory[T], section: string, value: T) =
    ## Adds value to the end of memory in section.
    if not (section in memory.sects):
        memory.sects[section] = memory.chunks.len
        memory.chunks.add @[]

    memory.chunks[memory.sects[section]].add value

proc add*[T](memory: var Memory[T], section: string, values: openarray[T]) =
    ## Adds value to the end of memory in section.
    if not (section in memory.sects):
        memory.sects[section] = memory.chunks.len
        memory.chunks.add @[]

    memory.chunks[memory.sects[section]].add values

proc address*[T](memory: Memory[T], section: string, offset: int): int =
    ## Computes the "real" address of a given section and offset
    return van_der_corput(memory.sects[section]) + offset

proc current*[T](memory: Memory[T], section: string): int =
    ## Computes the "real" address of the "next" element in the section
    let i = memory.sects[section]
    return van_der_corput(i) + len(memory.chunks[i])

proc addrinfo[T](memory: Memory[T], address: int): (int, int) =
    ## Calculates which section index an address points to and the offset in it
    var
        index = 0
        start = van_der_corput(index)

    for i in 1 .. memory.chunks.len:
        let corput = van_der_corput(i)
        if (corput < address) and (corput > start):
            index = i
            start = corput

    return (index, address - start)

proc contains*[T](memory: Memory[T], address: int): bool =
    ## Checks if the given address is a valid allocated address
    if address < SML_BLOCK_START:
        let (idx, off) = memory.addrinfo(address)
        return memory.chunks[idx].len >= off
    elif address < MID_BLOCK_START:
        let (blkid, offset) = divmod(address - SML_BLOCK_START, SML_BLOCK_SIZE)
        return memory.smlmem[blkid].len >= offset
    elif address < BIG_BLOCK_START:
        let (blkid, offset) = divmod(address - MID_BLOCK_START, MID_BLOCK_SIZE)
        return memory.midmem[blkid].len >= offset
    elif address < HUG_BLOCK_START:
        let (blkid, offset) = divmod(address - BIG_BLOCK_START, BIG_BLOCK_SIZE)
        return memory.bigmem[blkid].len >= offset
    else:
        let (blkid, offset) = divmod(address - HUG_BLOCK_START, HUG_BLOCK_SIZE)
        return memory.hugmem[blkid].len >= offset

proc `[]`*[T](memory: Memory[T], address: int): T =
    ## Accesses the underlying memory item for the given address, failing if it
    ## does not exist
    if address < SML_BLOCK_START:
        let (idx, off) = memory.addrinfo(address)
        return memory.chunks[idx][off]
    elif address < MID_BLOCK_START:
        let (blkid, offset) = divmod(address - SML_BLOCK_START, SML_BLOCK_SIZE)
        return memory.smlmem[blkid][offset]
    elif address < BIG_BLOCK_START:
        let (blkid, offset) = divmod(address - MID_BLOCK_START, MID_BLOCK_SIZE)
        return memory.midmem[blkid][offset]
    elif address < HUG_BLOCK_START:
        let (blkid, offset) = divmod(address - BIG_BLOCK_START, BIG_BLOCK_SIZE)
        return memory.bigmem[blkid][offset]
    else:
        let (blkid, offset) = divmod(address - HUG_BLOCK_START, HUG_BLOCK_SIZE)
        return memory.hugmem[blkid][offset]

proc `[]`*[T](memory: var Memory[T], address: int): var T =
    ## Accesses the underlying memory item for the given address, failing if it
    ## does not exist
    if address < SML_BLOCK_START:
        let (idx, off) = memory.addrinfo(address)
        return memory.chunks[idx][off]
    elif address < MID_BLOCK_START:
        let (blkid, offset) = divmod(address - SML_BLOCK_START, SML_BLOCK_SIZE)
        return memory.smlmem[blkid][offset]
    elif address < BIG_BLOCK_START:
        let (blkid, offset) = divmod(address - MID_BLOCK_START, MID_BLOCK_SIZE)
        return memory.midmem[blkid][offset]
    elif address < HUG_BLOCK_START:
        let (blkid, offset) = divmod(address - BIG_BLOCK_START, BIG_BLOCK_SIZE)
        return memory.bigmem[blkid][offset]
    else:
        let (blkid, offset) = divmod(address - HUG_BLOCK_START, HUG_BLOCK_SIZE)
        return memory.hugmem[blkid][offset]

proc `[]=`*[T](memory: var Memory[T], address: int, value: T) =
    ## Writes to the underlying memory item for a given address, failing if it
    ## does not exist
    if address < SML_BLOCK_START:
        let (idx, off) = memory.addrinfo(address)
        memory.chunks[idx][off] = value
    elif address < MID_BLOCK_START:
        let (blkid, offset) = divmod(address - SML_BLOCK_START, SML_BLOCK_SIZE)
        memory.smlmem[blkid][offset] = value
    elif address < BIG_BLOCK_START:
        let (blkid, offset) = divmod(address - MID_BLOCK_START, MID_BLOCK_SIZE)
        memory.midmem[blkid][offset] = value
    elif address < HUG_BLOCK_START:
        let (blkid, offset) = divmod(address - BIG_BLOCK_START, BIG_BLOCK_SIZE)
        memory.bigmem[blkid][offset] = value
    else:
        let (blkid, offset) = divmod(address - HUG_BLOCK_START, HUG_BLOCK_SIZE)
        memory.hugmem[blkid][offset] = value

proc write*[T](memory: var Memory[T], address: int, values: openArray[T]) =
    ## Writes to the underlying memory item for a given address, failing if it
    ## does not exist
    var a = address
    for i in values:
        memory[a] = i
        inc a

proc section*[T](memory: Memory[T], address: int): string =
    ## Calculates which section a given address points to (warning: slow)
    let (idx, off) = memory.addrinfo(address)

    for (section, index) in memory.sects.pairs:
        if index == idx:
            return section

    return ""

iterator sections*[T](memory: Memory[T]): string =
    for section in memory.sects.keys:
        yield section

proc data*[T](memory: Memory[T], section: string): seq[T] =
    return memory.chunks[memory.sects[section]]

proc allocate*[T](memory: var Memory[T], count: int): int =
    ## Allocates a block of count values that can be addressed starting from the
    ## returned value
    if count < SML_BLOCK_SIZE:
        var i = 0
        while i in memory.smlmem:
            inc i
        if i > SML_BLOCK_COUNT:
            raise new ResourceExhaustedError
        memory.smlmem[i] = newSeq[T](count)
        return SML_BLOCK_START + SML_BLOCK_SIZE * i
    elif count < MID_BLOCK_SIZE:
        var i = 0
        while i in memory.smlmem:
            inc i
        if i > MID_BLOCK_COUNT:
            raise new ResourceExhaustedError
        memory.smlmem[i] = newSeq[T](count)
        return MID_BLOCK_START + MID_BLOCK_SIZE * i
    elif count < BIG_BLOCK_SIZE:
        var i: int = 0
        while i in memory.smlmem:
            inc i
        if i > BIG_BLOCK_COUNT:
            raise new ResourceExhaustedError
        memory.smlmem[i] = newSeq[T](count)
        return BIG_BLOCK_START + BIG_BLOCK_SIZE * i
    else:
        var i = 0
        while i in memory.smlmem:
            inc i
        if i > HUG_BLOCK_COUNT:
            raise new ResourceExhaustedError
        memory.smlmem[i] = newSeq[T](count)
        return HUG_BLOCK_START + HUG_BLOCK_SIZE * i

proc deallocate*[T](memory: var Memory[T], address: int) =
    ## Deallocates the block containing address
    if address < SML_BLOCK_START:
        discard # Can't deallocate static memory
    elif address < MID_BLOCK_START:
        let blkid = (address - SML_BLOCK_START) div SML_BLOCK_SIZE
        memory.smlmem[blkid].setLen(0)
        memory.smlmem.del(blkid)
    elif address < BIG_BLOCK_START:
        let blkid = (address - MID_BLOCK_START) div MID_BLOCK_SIZE
        memory.midmem[blkid].setLen(0)
        memory.midmem.del(blkid)
    elif address < HUG_BLOCK_START:
        let blkid = (address - BIG_BLOCK_START) div BIG_BLOCK_SIZE
        memory.bigmem[blkid].setLen(0)
        memory.bigmem.del(blkid)
    else:
        let blkid = (address - HUG_BLOCK_START) div HUG_BLOCK_SIZE
        memory.hugmem[blkid].setLen(0)
        memory.hugmem.del(blkid)

proc reallocate*[T](memory: var Memory[T], address: int, count: int): int =
    ## Reallocates the block containing address to store up to count items
    var theseq: seq[T]
    if address < SML_BLOCK_START:
        return address # Can't reallocate static memory
    elif address < MID_BLOCK_START:
        let blkid = (address - SML_BLOCK_START) div SML_BLOCK_SIZE
        if count <= SML_BLOCK_SIZE:
            memory.smlmem[blkid].setLen(count)
            return address
        assert memory.smlmem.pop(blkid, theseq)
    elif address < BIG_BLOCK_START:
        let blkid = (address - MID_BLOCK_START) div MID_BLOCK_SIZE
        if count <= MID_BLOCK_SIZE:
            memory.midmem[blkid].setLen(count)
            return address
        assert memory.midmem.pop(blkid, theseq)
    elif address < HUG_BLOCK_START:
        let blkid = (address - BIG_BLOCK_START) div BIG_BLOCK_SIZE
        if count <= BIG_BLOCK_SIZE:
            memory.bigmem[blkid].setLen(count)
            return address
        assert memory.bigmem.pop(blkid, theseq)
    else:
        let blkid = (address - HUG_BLOCK_START) div HUG_BLOCK_SIZE
        if count <= HUG_BLOCK_SIZE:
            memory.hugmem[blkid].setLen(count)
            return address
        assert memory.hugmem.pop(blkid, theseq)

    theseq.setLen(count)

    if count < SML_BLOCK_SIZE:
        var i = 0
        while i in memory.smlmem:
            inc i
        if i > SML_BLOCK_COUNT:
            raise new ResourceExhaustedError
        memory.smlmem[i] = theseq
        return SML_BLOCK_START + SML_BLOCK_SIZE * i
    elif count < MID_BLOCK_SIZE:
        var i = 0
        while i in memory.smlmem:
            inc i
        if i > MID_BLOCK_COUNT:
            raise new ResourceExhaustedError
        memory.smlmem[i] = theseq
        return MID_BLOCK_START + MID_BLOCK_SIZE * i
    elif count < BIG_BLOCK_SIZE:
        var i = 0
        while i in memory.smlmem:
            inc i
        if i > BIG_BLOCK_COUNT:
            raise new ResourceExhaustedError
        memory.smlmem[i] = theseq
        return BIG_BLOCK_START + BIG_BLOCK_SIZE * i
    else:
        var i = 0
        while i in memory.smlmem:
            inc i
        if i > HUG_BLOCK_COUNT:
            raise new ResourceExhaustedError
        memory.smlmem[i] = theseq
        return HUG_BLOCK_START + HUG_BLOCK_SIZE * i
