import std/[critbits, strformat, tables]

import memory

type
    ArcticPointer* = int64

    ArcticType* {.union.} = object
        b*: int8
        s*: int16
        w*: int32
        i*: int64
        u*: uint64
        d*: float64
        p*: ArcticPointer
        f*: ArcticPointer

    ArcticVariableIndex* = enum
        VARIABLE_A, VARIABLE_B, VARIABLE_C,
        VARIABLE_I, VARIABLE_N,
        VARIABLE_X, VARIABLE_Y, VARIABLE_Z

    ArcticImmediateKind* = enum
        PLAIN, VARIABLE, INTEGER, NUMBER, SYMBOL

    ArcticImmediate* = object
        case kind*: ArcticImmediateKind
            of PLAIN: nil
            of VARIABLE: v*: ArcticVariableIndex
            of INTEGER: i*: int64
            of NUMBER: n*: float64
            of SYMBOL: s*: string

    ArcticOperation* = object
        code*: char
        immediate*: ArcticImmediate
        bigendian*: bool

    ArcticStack* = seq[ArcticType]

    ArcticState* = object
        memory*: Memory[uint8]
        code*: Memory[ArcticOperation]
        symbols*: CritBitTree[ArcticPointer] # as name or section _name
        stack*: ArcticStack
        pc*: ArcticPointer # NOT negative!
        registers*: array[ArcticVariableIndex, ArcticType]
        dynmem*: Table[ArcticPointer, int] # pointer -> length

    ArcticStepResult* = enum
        CONTINUE, BREAKPOINT, EXIT, ERROR

    ArcticBuiltin* = proc (state: var ArcticState): ArcticStepResult {.nimcall.}

proc `$`*(variable: ArcticVariableIndex): string =
    case variable:
        of VARIABLE_A: return "A"
        of VARIABLE_B: return "B"
        of VARIABLE_C: return "C"
        of VARIABLE_I: return "I"
        of VARIABLE_N: return "N"
        of VARIABLE_X: return "X"
        of VARIABLE_Y: return "Y"
        of VARIABLE_Z: return "Z"

proc `$`*(op: ArcticOperation): string =
    result &= op.code

    case op.immediate.kind:
        of PLAIN: discard
        of VARIABLE: result &= $op.immediate.v
        of INTEGER: result &= $op.immediate.i
        of NUMBER: result &= $op.immediate.n
        of SYMBOL: result &= $op.immediate.s

proc `$`*(state: ArcticState): string =
    result &= "Registers:\n"
    for idx in ArcticVariableIndex:
        result &= &" {idx} = {state.registers[idx]}\n"

    result &= "Stack:\n"

    for item in state.stack:
        result &= &" {item}\n"

    result &= "Sections:\n"

    for section in state.code.sections:
        result &= "\n  section "
        result &= section
        result &= ":\n"
        for (i, m) in state.code.data(section).pairs:
            for (label, location) in state.symbols.pairs:
                if -location == state.code.address(section, i):
                    result &= &":{label} "
                if state.pc == state.code.address(section, i):
                    result &= ">"
            if m.code != ' ':
                result &= $m

    for section in state.memory.sections:
        result &= "\n  section "
        result &= section
        result &= ":\n"
        for (i, m) in state.memory.data(section).pairs:
            for (label, location) in state.symbols.pairs:
                if location == state.memory.address(section, i):
                    result &= &":{label} "
            result &= &"{m} "

